/*********************************************************************
 * webcam_server.c 0.30                                              *
 *                                                                   *
 * (c) 2002 Donn Morrison dmorriso@gulf.uvic.ca                      *
 *                                                                   *
 * code used from Gerd Knorr's xawtv (libng)                         *
 * - and -                                                           *
 * Cory Lueninghoener's gqcam                                        *
 *                                                                   *
 *    waits for connections from a viewer and sends                  *
 *    jpeg encoded captures as a live video feed                     *
 *                                                                   *
 *********************************************************************/

#include <stdio.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <arpa/inet.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <linux/videodev.h>
#include <semaphore.h>
#include <pthread.h>
#include <jpeglib.h>
#include <jerror.h>

#include "webcam_server.h"

#define VERSION			"0.30"

#define MAX_READ_LEN		1024	/* max client socket read length (bytes) */
#define READ_TIMEOUT		15	/* client socket read timeout before disconnect (seconds) */
#define MAX_BAD_GRABS		10	/* max bad image captures before exit */
#define MAX_WATCHDOG_RETRIES	5	/* max watchdog retries */
#define MAX_OLD_IMAGES		20	/* max old images before disconnect */
#define DEFAULT_JPEG_SIZE	20000	/* default jpeg size (bytes) */
#define CAPTURE_PAUSE		75000	/*  pause for 75ms between captures */

/*-----------------------------------------------------------
  default options
  -----------------------------------------------------------*/
#define DEF_PORT		8888
#define DEF_QUALITY		60
#define DEF_CAMDEVFILE		"/dev/video0"
#define DEF_LOGFILE		"/var/log/webcam_server.log"
#define DEF_CAPTION		"%Y-%m-%d %H:%M:%S" /* strftime */
#define DEF_WIDTH		320
#define DEF_HEIGHT		240
#define DEF_RETRYINIT		0
#define DEF_DAEMON		0
#define DEF_SWAPRGB		0
#define DEF_TESTFPS		0
#define DEF_USENG		1
#define DEF_FLIP_HORIZ		0
#define DEF_FLIP_VERT		0
#define DEF_GAMMA		0
#define DEF_TEXT_FG		255
#define DEF_TEXT_BG		0
#define DEF_ROTATE		0

/*-----------------------------------------------------------
  jpeg data buffer stuff
  -----------------------------------------------------------*/
METHODDEF(long)
get_jpeg_size (j_compress_ptr cinfo);

typedef struct {
  struct jpeg_destination_mgr pub;

  long total_buffer_size;
  JOCTET * buffer;
} my_destination_mgr;

typedef my_destination_mgr * my_dest_ptr;

METHODDEF(void)
init_destination (j_compress_ptr cinfo)
{
  my_dest_ptr dest = (my_dest_ptr) cinfo->dest;

  dest->buffer = (char *)malloc(DEFAULT_JPEG_SIZE * sizeof(JOCTET));
  
  dest->pub.next_output_byte = dest->buffer;
  dest->pub.free_in_buffer = DEFAULT_JPEG_SIZE;
  dest->total_buffer_size = DEFAULT_JPEG_SIZE;
}

METHODDEF(boolean)
empty_output_buffer (j_compress_ptr cinfo)
{
  my_dest_ptr dest = (my_dest_ptr) cinfo->dest;
  u_char *tmp;
  long size;

  size = dest->total_buffer_size;

  /* re-allocate more memory (another DEFAULT_JPEG_SIZE block) */
  tmp = (char *)realloc(dest->buffer, (dest->total_buffer_size + DEFAULT_JPEG_SIZE) * sizeof(JOCTET));
  if(tmp == NULL)
  {
    fprintf(stderr, "empty_output_buffer: couldn't allocate more memory for jpeg (%d bytes)\n",
            (dest->total_buffer_size + DEFAULT_JPEG_SIZE) * sizeof(JOCTET));
    return FALSE;
  }
  dest->total_buffer_size += DEFAULT_JPEG_SIZE;
  dest->buffer = tmp;
  /* reset the free_in_buffer counter */
  dest->pub.free_in_buffer = dest->total_buffer_size - size;
  dest->pub.next_output_byte = dest->buffer + size;
  return TRUE;
}

METHODDEF(void)
term_destination (j_compress_ptr cinfo)
{
  my_dest_ptr dest = (my_dest_ptr) cinfo->dest;
}

GLOBAL(void)
jpeg_buff_dest (j_compress_ptr cinfo)
{
  my_dest_ptr dest;

  if (cinfo->dest == NULL) {	/* first time for this JPEG object? */
    cinfo->dest = (struct jpeg_destination_mgr *)
      (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
				  sizeof(my_destination_mgr));
  }

  dest = (my_dest_ptr) cinfo->dest;
  dest->pub.init_destination = init_destination;
  dest->pub.empty_output_buffer = empty_output_buffer;
  dest->pub.term_destination = term_destination;
}

METHODDEF(long)
get_jpeg_size (j_compress_ptr cinfo)
{
	my_dest_ptr dest = (my_dest_ptr) cinfo->dest;
	
	return (long)(dest->pub.next_output_byte - dest->buffer);
}

METHODDEF(void)
jpeg_buff_free (j_compress_ptr cinfo)
{
  my_dest_ptr dest = (my_dest_ptr) cinfo->dest;
  
  free(dest->buffer);
}

/*-----------------------------------------------------------
  grab_thread
  - runs only when clients are connected or when test mode
    enabled
  - captures an image, converts it to jpeg, pushes it on 
    queue, loops
  -----------------------------------------------------------*/
static int grab_thread(struct caminfo *cam)
{
	char *jpeg_data = NULL, buff[256];
	struct imagequeue *new;
	int len, width, height;
	int bad_grabs = 0;
	u_long frames=0;
	time_t t0;
	time_t t1;
	struct timeb start_time;
	struct timeb stop_time;
	unsigned int capture_time;

	if(!cam->o.use_ng)
	{
		if(cam->raw_image)
			free(cam->raw_image);
		cam->raw_image = (char *)malloc(cam->o.width * cam->o.height * 3);
		if(!cam->raw_image)
		{
			perror("malloc");
			return 1;
		}
	}

	time(&t0);
	while(cam->grab_thread_alive)
	{
		if(!cam->o.test_fps)
		{
			/* wait for a connection */
			sem_wait(&sem_con);
			sem_post(&sem_con);
		}

		ftime(&start_time);

		if(cam->o.use_ng)
		{
			/* read an image from camera using libng from xawtv */
			cam->raw_image = ng_grab_one(cam, &cam->o.width, &cam->o.height);

			if(!cam->raw_image || cam->o.width == 0 || cam->o.height == 0)
			{
				fprintf(stderr,"bad image grab! %d\n", bad_grabs);
				usleep(100000);
				bad_grabs++;
				continue;
			}
			len = cam->o.width * cam->o.height * 3;
		}
		else
		{
			/* read an image from camera */
			len = read(cam->dev, cam->raw_image, 
				   cam->o.width * cam->o.height * 3);

			if(len <= 0)
			{
				fprintf(stderr,"bad image grab! %d\n", bad_grabs);
				usleep(100000);
				bad_grabs++;
				continue;
			}
		}

		time(&t1);
		if(cam->o.test_fps)
		{
			float fps;
			/* test mode only */
			frames++;
			fps = (float)frames/((t1 - t0) ? t1-t0 : 1.0);
			printf("read a frame: %d bytes, fps = %.3f, frames = %d\n", len, fps, frames);
			continue;
		}

		/* swap rgb? */
		if(cam->o.swap_rgb)
			swap_rgb24(cam->raw_image, 
				   cam->o.width * cam->o.height);
		
		/* adjust gamma? */
		if(cam->o.gamma)
			adjust_gamma(cam->raw_image,
				     cam->o.width * cam->o.height * 3, cam->o.gamma);

		/* flip horiz/vert */
		if(cam->o.flip_horiz)
			fliph(cam->raw_image, cam->o.width, cam->o.height);
		if(cam->o.flip_vert)
			flipv(cam->raw_image, cam->o.width, cam->o.height);

		width = cam->o.width;
		height = cam->o.height;

		if(cam->o.rotate > 0)
			rotate_image(cam->raw_image, &width, &height,
				    cam->o.rotate, 0, 0, cam->o.height, cam->o.width);

		/* add the caption? */
		if(cam->o.caption)
			add_text(cam, 
				 cam->raw_image, 
				 width, 
				 height);
		
		/* allocate some storage for the new image */
		new = (struct imagequeue *)malloc(sizeof(struct imagequeue));
		if(!new)
		{
			perror("malloc");
			clean(cam);
			continue;
		}
		
		/* compress the image in jpeg format */
		len = raw_to_jpeg(cam->raw_image,
				  &jpeg_data, 
				  width, 
				  height, 
				  cam->o.jpeg_quality);

		pthread_mutex_lock(&lock_queue);
		pthread_mutex_lock(&lock_id);

		/* set fields */
		new->id = ++(cam->s.id_newest_img);
		new->len = len;
		new->jpeg_data = jpeg_data;
		new->ref = 0;
		new->next = NULL;
		pthread_mutex_init(&new->lock_ref, NULL);
	
		/* put it on the queue */
		push(cam, new);
		
		pthread_mutex_unlock(&lock_id);

		ftime(&stop_time);
		capture_time = (stop_time.time - start_time.time);
		if (capture_time == 0)
		{
			capture_time = stop_time.millitm - start_time.millitm;
		}
		else
		{
			capture_time *= 1000;
			capture_time += (stop_time.millitm - start_time.millitm);
		}
		capture_time *= 100;
		if (capture_time < CAPTURE_PAUSE)
			usleep(CAPTURE_PAUSE - capture_time);

		/* clear unused images from the queue */
		len = clean(cam);
		pthread_mutex_unlock(&lock_queue);
	}
	
	fprintf(stderr,"grab thread exiting...\n");
	if(!cam->o.use_ng)
		free(cam->raw_image);
	exit(1);
}

/*-----------------------------------------------------------
  watchdog_thread
  - simulates a client connection and ensures the grabber
    thread is working. this is a quick hack to unstick
    dead read()s
  -----------------------------------------------------------*/
static int watchdog_thread(struct caminfo *cam)
{
	struct imagequeue *ptr;
	int last_id = -1;
	int tries = 0;
	int res;
	char buff[256];
	
	while(1)
	{
		/* simulate a client connection */
		sem_post(&sem_con);
		
		/* get the latest image */
		pthread_mutex_lock(&lock_queue);
		ptr = peek(cam);
		if(!ptr)
		{
			/* no images available */
			pthread_mutex_unlock(&lock_queue);
		}
		else
		{
			pthread_mutex_lock(&ptr->lock_ref);
			ptr->ref++;
			pthread_mutex_unlock(&ptr->lock_ref);

			pthread_mutex_unlock(&lock_queue);

			if(ptr->id == last_id)
			{
				/* hmm...same id as last time */
				tries++;
				sprintf(buff,"watchdog: image (%d) was the same as last time! %d/%d\n",
					last_id, tries, MAX_WATCHDOG_RETRIES);
				log(cam,buff);
			}
			else
			{
				last_id = ptr->id;
				tries = 0;
			}
			
			pthread_mutex_lock(&ptr->lock_ref);
			ptr->ref--;
			pthread_mutex_unlock(&ptr->lock_ref);
		}
		
		if(tries == MAX_WATCHDOG_RETRIES)
		{
			log(cam,"watchdog: ERROR: grabber thread suspected stuck\n");
			/* drastic action!
			 * we assume the grabber thread is stuck
			 * on a read, and we kill it, reinit the
			 * camera, and restart the grabber thread
			 */
			tries = 0;
			last_id = -1;
			
			log(cam,"watchdog: killing grabber thread\n");
			pthread_cancel(cam->pt_grab);
			
			log(cam,"watchdog: closing camera\n");
			close_cam(cam);
			
			log(cam,"watchdog: camera closed\n");
			usleep(5000000);
			
			log(cam,"watchdog: re-initing camera\n");
			if(cam->o.use_ng)
				res = ng_grab_init(cam);
			else
				res = open_cam(cam, cam->o.devfile);
			
			while(res < 0 && cam->o.retry_init)
			{
				log(cam,"watchdog: re-init failed, retrying\n");
				usleep(1000000);
				if(cam->o.use_ng)
					res = ng_grab_init(cam);
				else
					res = open_cam(cam, cam->o.devfile);
			}

			if(res < 0)
			{
				log(cam,"watchdog: re-init failed, exiting\n");
				exit(1);
			}

			sprintf(buff, "watchdog: re-init ok: %d\n", res);
			log(cam,buff);
			
			log(cam,"watchdog: setting camera defaults\n");
			if(!cam->o.use_ng)
			{
				/* get camera defaults */
				get_cam_info(cam);
				cam->vid_win.width = cam->o.width;
				cam->vid_win.height = cam->o.height;
				if(set_cam_info(cam) < 0)
				{
					get_cam_info(cam);
					if(set_cam_info(cam) < 0)
					{
						exit(1);
					}
				}
				cam->o.width = cam->vid_win.width;
				cam->o.height = cam->vid_win.height;
			}
			
			log(cam,"watchdog: restarting grab_thread\n");
			cam->grab_thread_alive = 1;
			pthread_create(&cam->pt_grab, NULL, (void*)grab_thread, (void *)cam);
			pthread_detach(cam->pt_grab);
		}
		
		sem_wait(&sem_con);
		
		/* sleep and then check again */
		usleep(2000000);
	}
}

/*-----------------------------------------------------------
  handle_connection
  - waits for a client request, gets and sends latest
    image from queue, loops
  - does some connection accounting
  -----------------------------------------------------------*/
static int handle_connection(struct connection *con)
{
	struct caminfo *cam = con->cam;
	struct imagequeue *ptr;
	long len, res;
	char read_buff[MAX_READ_LEN];
	char logmsg[512];
	int got_request = 0;

	/* get start time */
	time(&con->start);
	pthread_mutex_lock(&lock_info);
	cam->s.num_users++;
	pthread_mutex_unlock(&lock_info);

	sprintf(logmsg, "%s connected\n", inet_ntoa(con->remote_addr.sin_addr));
	log(cam, logmsg);

	/* someone has connected, inform the grabber thread */
	sem_post(&sem_con);

	con->total_bytes = 0;
	con->total_frames = 0;
	con->id_last = 0;
	con->old_images = 0;

	while(1)
	{
		res = 1;

		if(!got_request)
		{
			struct timeval tv;
			fd_set rfds;
			tv.tv_sec = READ_TIMEOUT;
			tv.tv_usec = 0;

			FD_ZERO(&rfds);
			FD_SET(con->socketfd, &rfds);

			/* wait for client request */
			if(!select(con->socketfd+1, &rfds, 0, 0, &tv) || !FD_ISSET(con->socketfd, &rfds))
				break;

			len = recv(con->socketfd, &read_buff, MAX_READ_LEN, 0);
			read_buff[len] = 0;

			if(len < 1)
				break;
			got_request = 1;
		}

		/* get the latest image */
		pthread_mutex_lock(&lock_queue);
		ptr = peek(cam);
		if(!ptr)
		{
			pthread_mutex_unlock(&lock_queue);
			usleep(100000);
			continue;
		}
		pthread_mutex_lock(&ptr->lock_ref);
		ptr->ref++;
		pthread_mutex_unlock(&ptr->lock_ref);

		pthread_mutex_unlock(&lock_queue);

		/* don't send the same image more than once */
		if(ptr->id != con->id_last)
		{
			/* send HTTP OK n stuff */
			sprintf(read_buff, "HTTP/1.0 200 OK\nContent-type: image/jpeg\nContent-Length: %d\nConnection: close\n\n", ptr->len);
			if(send(con->socketfd, read_buff, strlen(read_buff),0) != strlen(read_buff))
			{
				res = 0;
			}

			/* send jpeg */
			if(res && send(con->socketfd, ptr->jpeg_data, ptr->len, 0) != ptr->len)
			{
				res = 0;
			}

			con->total_bytes += ptr->len;
			con->total_frames++;
			con->id_last = ptr->id;
			con->old_images = 0;
			got_request = 0;
		}
		else
		{
			/* no new image...wait a while and try again */
			usleep(100000);
			con->old_images++;
			if(con->old_images == MAX_OLD_IMAGES)
			{
				/* too many old images */
				pthread_mutex_lock(&ptr->lock_ref);
				ptr->ref--;
				pthread_mutex_unlock(&ptr->lock_ref);
				break;
			}
		}

		pthread_mutex_lock(&ptr->lock_ref);
		ptr->ref--;
		pthread_mutex_unlock(&ptr->lock_ref);

		if(!res)
			break;
	}

	/* get end time */
	time(&con->end);
	
	sprintf(logmsg, "%s disconnected, %d seconds, %d bytes, %d Kbytes/second, %d frames, %d fps\n",
		inet_ntoa(con->remote_addr.sin_addr),		/* address */
		con->end - con->start,				/* seconds */
		con->total_bytes,				/* bytes */
		(con->total_bytes/1024)/((con->end - con->start) ? con->end - con->start : 1),/* Kbytes/sec */
		con->total_frames,				/* frames */
		con->total_frames/((con->end - con->start) ? con->end - con->start : 1));	/* frames/sec */
	log(cam,logmsg);
	
	pthread_mutex_lock(&lock_info);
	cam->s.num_users--;
	pthread_mutex_unlock(&lock_info);
	
	/* someone has disconnected, inform the grabber thread */
	sem_wait(&sem_con);

	close(con->socketfd);
	free(con);
	return 0;
}

/*-----------------------------------------------------------
  imagequeue procedures
  
  push: appends a fresh image to the queue
  clean: removes unreferenced (old) images
  peek: retrieves the newest image in the queue
        without removing it
  -----------------------------------------------------------*/
void push(struct caminfo *cam, struct imagequeue *new)
{
	struct imagequeue *ptr;

	ptr = cam->queue;
	while(ptr)
	{
		if(ptr->next == NULL)
			break;
		ptr = ptr->next;
	}

	if(ptr)
	{
		ptr->next = new;
		new->prev = ptr;
	}
	else
	{
		cam->queue = new;
		new->prev = NULL;
	}
}

int clean(struct caminfo *cam)
{
	struct imagequeue *ptr, *tmp;
	int count = 0;

	ptr = cam->queue;
	while(ptr)
	{
		tmp = ptr->next;
		/* only remove it if its not being used and its not the latest */
		pthread_mutex_lock(&ptr->lock_ref);
		if(ptr->ref == 0 && ptr->id != cam->s.id_newest_img)
		{
			if(ptr == cam->queue)
				cam->queue = tmp;
			if(ptr->prev)
				ptr->prev->next = ptr->next;
			if(ptr->next)
				ptr->next->prev = ptr->prev;
			free(ptr->jpeg_data);
			pthread_mutex_unlock(&ptr->lock_ref);
			pthread_mutex_destroy(&ptr->lock_ref);
			free(ptr);
			count++;
		}
		else
			pthread_mutex_unlock(&ptr->lock_ref);
		
		ptr = tmp;
	}

	return count;
}

struct imagequeue *peek(struct caminfo *cam)
{
	struct imagequeue *ptr;
	int count = 0, id;

	pthread_mutex_lock(&lock_id);
	id = cam->s.id_newest_img;
	pthread_mutex_unlock(&lock_id);

	ptr = cam->queue;
	while(ptr)
	{
		if(ptr->id == id)
			break;
		ptr = ptr->next;
		count++;
	}
	
	return ptr;
}

/*-----------------------------------------------------------
  image processing procedures
  
  raw_to_jpeg: encodes raw image to jpeg
  add_text: adds a caption to the image
  swap_rgb24: RGB->RBG
  fliph: flips the image horizontally
  flipv: flips vertically
  -----------------------------------------------------------*/
int raw_to_jpeg(char *raw_image, char **jpeg_data, int w, int h, int quality)
{
	struct jpeg_compress_struct cinfo;
	struct jpeg_error_mgr jerr;
	int i, size;
	unsigned char *line;

	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_compress(&cinfo);
	jpeg_buff_dest(&cinfo);
	cinfo.image_width = w;
	cinfo.image_height = h;
	cinfo.input_components = 3;
	cinfo.in_color_space = JCS_RGB;
	jpeg_set_defaults(&cinfo);
	jpeg_set_quality(&cinfo, quality, TRUE);
	jpeg_start_compress(&cinfo, TRUE);
	for (i = 0, line = raw_image; i < h; i++, line += w*3)
		jpeg_write_scanlines(&cinfo, &line, 1);
	jpeg_finish_compress(&cinfo);
	
	size = get_jpeg_size(&cinfo);
	
	/* this memory must be freed elsewhere */
	(*jpeg_data) = (char *)malloc(size);
	if(!(*jpeg_data))
	{
		jpeg_destroy_compress(&cinfo);
		jpeg_buff_free(&cinfo);
		return -1;
	}
	memcpy(*jpeg_data, ((my_dest_ptr)cinfo.dest)->buffer, size);

	jpeg_destroy_compress(&cinfo);
	jpeg_buff_free(&cinfo);
	
	return size;
}

/* Gerd Knorr */

#define CHAR_HEIGHT  11
#define CHAR_WIDTH   6
#define CHAR_START   4
#include "font_6x11.h"

void add_text(struct caminfo *cam, char *image, int width, int height)
{
	time_t      t;
	struct tm  *tm;
	char        *line,*ptr,prepend[32];
	int         i,x,y,f,len,size,numlines;

	if(cam->o.caption && strlen(cam->o.caption) < 1)
		return;

	pthread_mutex_lock(&lock_info);
	sprintf(prepend, ", [%d] %d viewer", cam->s.id_newest_img, cam->s.num_users);
	if(cam->s.num_users != 1)
	strcat(prepend, "s");
	pthread_mutex_unlock(&lock_info);

	size = strlen(cam->o.caption) +
	       strlen(prepend) +
	       128 +			/* 128 for strftime and null */
	       height/CHAR_HEIGHT;	/* total lines space in image */

	line = (char *)malloc(size);
	if(!line)
		return;

	time(&t);
	tm = localtime(&t);
	len = strftime(line,size-1,cam->o.caption,tm) + strlen(prepend);
	strcat(line,prepend);

	numlines = strlen(line) / (width/CHAR_WIDTH) + (strlen(line)%(width/CHAR_WIDTH) ? 1 : 0);

	for (y = 0; y < CHAR_HEIGHT; y++) {
		numlines = strlen(line) / (width/CHAR_WIDTH) + (strlen(line)%(width/CHAR_WIDTH) ? 1 : 0);
		ptr = image + 3 * width * (height-CHAR_HEIGHT-2 - ((numlines-1) * CHAR_HEIGHT) +y) + 12;
		for (x = 0; x < len; x++) {
			if(x % ((width/CHAR_WIDTH)-1) == 0 && x > 0)
			{
				numlines --;
				ptr = image + 3 * width * (height-CHAR_HEIGHT-2 - ((numlines-1) * CHAR_HEIGHT) +y) + 12;
			}
			f = fontdata[line[x] * CHAR_HEIGHT + y];
			for (i = CHAR_WIDTH-1; i >= 0; i--) {
				if (f & (CHAR_START << i)) {
					ptr[0] = cam->o.text_fg;
					ptr[1] = cam->o.text_fg;
					ptr[2] = cam->o.text_fg;
				}
				ptr += 3;
			}
		}
	}
	free(line);
}

void swap_rgb24(char *mem, int n)
{
        char  c;
        char *p = mem;
        int   i = n;

        while (--i) {
                c = p[0]; p[0] = p[2]; p[2] = c;
                p += 3;
        }
}

void adjust_gamma(char *mem, int n, int corr)
{
	int i=n;
	char *p = mem;
	while(--i)
	{
		if(corr > 0)
		{
			if(p[0] + corr <= 255)
				p[0] += corr;
			else
				p[0] = 255;
		}
		else
		{
			if(p[0] + corr >= 0)
				p[0] += corr;
			else
				p[0] = 0;
		}
		p++;
	}
}

static unsigned char *
rotate_image(unsigned char * in, int *wp, int *hp, int rot,
	     int top, int left, int bottom, int right)
{
    static unsigned char * rotimg = NULL;

    int i, j;

    int w = *wp;
    int ow = (right-left);
    int oh = (bottom-top);

    if (rotimg == NULL && (rotimg = (char*)malloc(ow*oh*3)) == NULL ) {
	fprintf(stderr, "out of memory\n");
	exit(1);
    }
    switch ( rot ) {
    default:
    case 0:
	for (j = 0; j < oh; j++) {
	    int ir = (j+top)*w+left;
	    int or = j*ow;
	    for (i = 0; i < ow; i++) {
		rotimg[3*(or + i)]   = in[3*(ir+i)];
		rotimg[3*(or + i)+1] = in[3*(ir+i)+1];
		rotimg[3*(or + i)+2] = in[3*(ir+i)+2];
	    }
	}
	*wp = ow;
	*hp = oh;
	break;
    case 1:
	for (i = 0; i < ow; i++) {
	    int rr = (ow-1-i)*oh;
	    int ic = i+left;
	    for (j = 0; j < oh; j++) {
		rotimg[3*(rr+j)]   = in[3*((j+top)*w+ic)];
		rotimg[3*(rr+j)+1] = in[3*((j+top)*w+ic)+1];
		rotimg[3*(rr+j)+2] = in[3*((j+top)*w+ic)+2];
	    }
	}
	*wp = oh;
	*hp = ow;
	break;
    case 2:
	for (j = 0; j < oh; j++) {
	    int ir = (j+top)*w;
	    for (i = 0; i < ow; i++) {
		rotimg[3*((oh-1-j)*ow + (ow-1-i))] = in[3*(ir+i+left)];
		rotimg[3*((oh-1-j)*ow + (ow-1-i))+1] = in[3*(ir+i+left)+1];
		rotimg[3*((oh-1-j)*ow + (ow-1-i))+2] = in[3*(ir+i+left)+2];
	    }
	}
	*wp = ow;
	*hp = oh;
	break;
    case 3:
	for (i = 0; i < ow; i++) {
	    int rr = i*oh;
	    int ic = i+left;
	    rr += oh-1;
	    for (j = 0; j < oh; j++) {
		rotimg[3*(rr-j)]   = in[3*((j+top)*w+ic)];
		rotimg[3*(rr-j)+1] = in[3*((j+top)*w+ic)+1];
		rotimg[3*(rr-j)+2] = in[3*((j+top)*w+ic)+2];
	    }
	}
	*wp = oh;
	*hp = ow;
	break;
    }
    memcpy(in, rotimg, ow*oh*3);
    return rotimg;
}

static unsigned int
compare_images(unsigned char *last, unsigned char *current,
	       int width, int height)
{
    unsigned char *p1 = last;
    unsigned char *p2 = current;
    int avg, diff, max, i = width*height*3;

    for (max = 0, avg = 0; --i; p1++,p2++) {
	diff = (*p1 < *p2) ? (*p2 - *p1) : (*p1 - *p2);
	avg += diff;
	if (diff > max)
	    max = diff;
    }
    avg = avg / width / height;
    fprintf(stderr,"compare: max=%d,avg=%d\n",max,avg);
    /* return avg */
    return max;
}

void fliph(char *image, int width, int height)
{
	int i,j;
	char *dest = NULL;
	char *srcline, *destline;

	dest = (char *)malloc(width*height*3);
	if(!dest)
		return;

	for(i=0;i<height;i++)
	{
		srcline = image + i*3*width;
		destline = dest + i*3*width + 3*(width-1);
		for(j=0;j<width;j++)
		{
			memcpy(destline, srcline, 3);
			srcline += 3;
			destline -= 3;
		}
	}

	memcpy(image, dest, width*height*3);
	free(dest);
}

void flipv(char *image, int width, int height)
{
	int i,j;
	char *dest = NULL;
	char *srcline, *destline;
	
	dest = (char *)malloc(width*height*3);
	if(!dest)
		return;
	
	for(i=0;i<height;i++)
	{
		srcline = image + i*width*3;
		destline = dest + (height-i)*width*3 - width*3;
	//	printf("%d - %d = %d\n", height*width*3, (height-i)*width*3 - width*3, height*width*3 - ((height-i)*width*3 - width*3));
		memcpy(destline, srcline, width*3);
	}
	
	memcpy(image, dest, width*height*3);
	free(dest);
}

/*-----------------------------------------------------------
  camera procedures
  
  open_cam: attempts to open a camera's device file
  close_cam: pretty obvious
  get_cam_info/set_cam_info: get/set camera attributes
  -----------------------------------------------------------*/
int open_cam(struct caminfo *cam, char *devfile)
{
	if(!devfile)
		return -1;
	/* initialize and open camera */
	cam->dev = open(devfile, O_RDONLY/* | O_NONBLOCK*/);
	if(cam->dev < 0)
	{
		char msg[256];
		sprintf(msg, "%s", devfile);
		perror(msg);
	}
	return cam->dev;
}

int close_cam(struct caminfo *cam)
{
	return close(cam->dev);
}

int set_cam_info(struct caminfo *cam)
{
	if (ioctl (cam->dev, VIDIOCSPICT, &cam->vid_pic) == -1) {
		perror ("ioctl (VIDIOCSPICT)");
		return -1;
	}
	if (ioctl (cam->dev, VIDIOCSWIN, &cam->vid_win) == -1) {
		perror ("ioctl (VIDIOCSWIN)");
		return -1;
	}
	return 0;
}

void get_cam_info(struct caminfo *cam)
{
	ioctl(cam->dev, VIDIOCGCAP, &cam->vid_caps);
	ioctl(cam->dev, VIDIOCGWIN, &cam->vid_win);
	ioctl(cam->dev, VIDIOCGPICT, &cam->vid_pic);
}

int read_timeout(int fd, char *dest, int size, int seconds)
{
	time_t t0,t1;
	int res,i=0;
	
	while(1)
	{
		time(&t0);
		while((res = read(fd, &dest[i], 1)) == 1)
		{
			i++;
			if(i == size)
				break;
			time(&t0);
		}
		time(&t1);
		if((t1 - t0) > seconds)
			break;
	}	 
	
	return res > 0 ? i : res;
}

/*-----------------------------------------------------------
  libng stuff (from xawtv)
  
  ng_grab_init: initialise ng grabber device
  ng_grab_one: grabs an image
  -----------------------------------------------------------*/
int ng_grab_init(struct caminfo *cam)
{
	struct ng_attribute *attr;
	int val,i;
	ng_init();

	ng_dev.video= cam->o.devfile;
	cam->ng_drv = ng_vid_open(ng_dev.video,NULL,0,&cam->ng_h_drv);
	if (NULL == cam->ng_drv) {
		fprintf(stderr,"no grabber device available\n");
		return -1;
	}
	if (!(cam->ng_drv->capabilities(cam->ng_h_drv) & CAN_CAPTURE)) {
		fprintf(stderr,"device doesn't support capture\n");
		return -1;
	}

	/* try native */
	cam->ng_fmt.fmtid  = VIDEO_RGB24;
	cam->ng_fmt.width  = cam->o.width;
	cam->ng_fmt.height = cam->o.height;
	if (0 == cam->ng_drv->setformat(cam->ng_h_drv,&cam->ng_fmt))
		return 0;

	/* check all available conversion functions */
	cam->ng_fmt.bytesperline = cam->ng_fmt.width*ng_vfmt_to_depth[cam->ng_fmt.fmtid]/8;
	for (i = 0;;) {
		cam->ng_conv = ng_conv_find(cam->ng_fmt.fmtid, &i);
		if (NULL == cam->ng_conv)
			break;
		cam->ng_gfmt = cam->ng_fmt;
		cam->ng_gfmt.fmtid = cam->ng_conv->fmtid_in;
		cam->ng_gfmt.bytesperline = 0;
		if (0 == cam->ng_drv->setformat(cam->ng_h_drv,&cam->ng_gfmt)) {
			cam->ng_fmt.width  = cam->ng_gfmt.width;
			cam->ng_fmt.height = cam->ng_gfmt.height;
			cam->ng_hconv = cam->ng_conv->init(&cam->ng_fmt,cam->ng_conv->priv);
			return 0;
		}
	}
	fprintf(stderr,"can't get rgb24 data\n");
	return -1;
}

char* ng_grab_one(struct caminfo *cam, int *width, int *height)
{
	static struct ng_video_buf *cap,*buf;

	if (NULL != buf)
		ng_release_video_buf(buf);
	if (NULL == (cap = cam->ng_drv->getimage(cam->ng_h_drv))) {
		fprintf(stderr,"capturing image failed\n");
		return NULL;
	}

	if (NULL != cam->ng_conv) {
		buf = ng_malloc_video_buf(&cam->ng_fmt,3*cam->ng_fmt.width*cam->ng_fmt.height);
		cam->ng_conv->frame(cam->ng_hconv,buf,cap);
		buf->info = cap->info;
		ng_release_video_buf(cap);
	} else {
		buf = cap;
	}

	*width  = buf->fmt.width;
	*height = buf->fmt.height;
	return buf->data;
}

/*-----------------------------------------------------------
  create_and_listen
  - creates a server socket and listens
  - its a seperate function to clean up main
  -----------------------------------------------------------*/
int create_and_listen(int port)
{
	int sockfd, yes=1;
	struct sockaddr_in addr;
	
	if((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		perror("socket");
		return -1;
	}
	
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	
	if(setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) < 0)
	{
		perror("setsockopt");
		return -1;
	}
	
	if(bind(sockfd, (struct sockaddr *)&addr, sizeof(struct sockaddr)) < 0)
	{
		perror("bind");
		return -1;
	}
	
	if(listen(sockfd, 5) < 0)
	{
		perror("listen");
		return -1;
	}
	
	return sockfd;
}

/*-----------------------------------------------------------
  log
  - opens a logfile for appending on first call
  - subsequent calls write to the open logfile and/or stdout
  -----------------------------------------------------------*/
void log(struct caminfo *cam, char *text)
{
	static int tried = 0;
	static int dups = 0;
	time_t      t;
	struct tm  *tm;
	char date[128];
	static char *last_msg = NULL;

	if(!tried)
	{
		/* open log file */
		cam->log = fopen(cam->o.logfile, "a");
		if(cam->log == NULL)
		{
			fprintf(stderr, "could not open logfile %s\n", cam->o.logfile);
		}
		tried = 1;
	}
	
	time(&t);
	tm = localtime(&t);
	strftime(date,127,"%Y-%m-%d %H:%M:%S",tm);
	
	if(last_msg)
	{
		if(strcmp(last_msg, text) == 0)
		{
			dups++;
			return;
		}
		else
		{
			free(last_msg);
			last_msg = NULL;
		}
	}

	/* write the message to stdout and/or logfile */
	pthread_mutex_lock(&lock_log);
	if(cam->o.verbose)
	{
		if(dups > 0)
		{
			fprintf(stdout, "%s [%s] log: the last message repeated %d times\n", date, cam->o.devfile, dups);
		}
		
		fprintf(stdout, "%s [%s] %s", date, cam->o.devfile, text);
		fflush(stdout);
	}
		
	if(cam->log)
	{
		if(dups > 0)
		{
			fprintf(cam->log, "%s [%s] log: the last message repeated %d times\n", date, cam->o.devfile, dups);
		}
		fprintf(cam->log, "%s [%s] %s", date, cam->o.devfile, text);
		fflush(cam->log);
	}
	pthread_mutex_unlock(&lock_log);
	
	if(!last_msg)
	{
		last_msg = (char *)malloc(strlen(text)+1);
		strcpy(last_msg, text);
		dups = 0;
	}
}

/*-----------------------------------------------------------
  usage, parse_args
  - highly complex, undocumented functions used for encrypting
    and secretly displaying or parsing various portions of text
  -----------------------------------------------------------*/
void usage()
{
printf("webcam_server %s, copyright 2001,2002 donn morrison\n\n", VERSION);
printf("usage: webcam_server [options]\n\n\
     option             description [default value]\n\n\
     -d video_device    - v4l compliant device [/dev/video]\n\
     -p tcp_port        - tcp listen port [8888]\n\
     -c \"caption\"       - caption text [%%Y-%%m-%%d %%H:%%M:%%S]\n\
     -tf colour         - text forground colour (0-255) [255]\n\
     -tb colour         - text background colour (0-255) [0]\n\
     -q jpeg_quality    - jpeg compression quality (0-100) [60]\n\
     -g geometry        - change default geometry (WIDTHxHEIGHT) [320x240]\n\
     -x                 - swap RGB -> BGR [no]\n\
     -G gamma		- gamma correction (-100-100) [0]\n\
     -fh		- flip image horizontally\n\
     -fv		- flip image vertically\n\
     -R value           - rotate image value*90 degrees (0-3) [0]\n\
     -n                 - don't use libng (recommended for some cameras!!) [no]\n\
     -r                 - if init fails, retry forever (hotplugging) [no]\n\
     -s                 - daemon mode [no]\n\
     -v                 - verbose [no]\n\
     -l logfile         - log to file [/var/log/webcam_server.log]\n\
     -a                 - test mode (fps) [no]\n\
\n");
}

int parse_args(struct caminfo *cam, int argc, char *argv[])
{
	int res = 0, i;
	for(i=1; i<argc; i++)
	{
		if(argv[i][0] == '-')
		{
			switch(argv[i][1])
			{
			case 'h':
				usage();
				return -1;
			case 'p':
				if(i+1 < argc)
					cam->o.port = atoi(argv[++i]);
				else
					res = -1;
				break;
			case 'd':
				if(i+1 < argc)
					cam->o.devfile = argv[++i];
				else
					res = -1;
				break;
			case 'c':
				if(i+1 < argc)
					cam->o.caption = argv[++i];
				else
					res = -1;
				break;
			case 'v':
				cam->o.verbose = 1;
				break;

			case 'r':
				cam->o.retry_init = 1;
				break;
			case 's':
				cam->o.daemon = 1;
				break;
			case 'q':
				if(i+1 < argc)
					cam->o.jpeg_quality = atoi(argv[++i]);
				else
					res = -1;
				break;
			case 'x':
				cam->o.swap_rgb = 1;
				break;
			case 'f':
				switch(argv[i][2])
				{
					case 'h':
						cam->o.flip_horiz = 1;
						break;
					case 'v':
						cam->o.flip_vert = 1;
						break;
				}
				break;
			case 't':
				switch(argv[i][2])
				{
					case 'f':
						cam->o.text_fg = atoi(argv[++i]);
						break;
					case 'b':
						cam->o.text_bg = atoi(argv[++i]);
						break;
				}
				break;
			case 'a':
				cam->o.test_fps = 1;
				break;
			case 'n':
				cam->o.use_ng = 0;
				break;
			case 'l':
				if(i+1 < argc)
					cam->o.logfile = argv[++i];
				else
					res = -1;
				break;
			case 'R':
				if(i+1 < argc)
					cam->o.rotate = atoi(argv[++i]);
				else
					res = -1;
				break;
			case 'g':
				if(i+1 < argc)
				{
					int w,h;
					char *arg = argv[++i];
					char *ptr;
					ptr = (char *)strchr(arg, 'x');
					if(!ptr)
					{
						res = -1;
						break;
					}
					*ptr = 0;
					ptr++;
					w = atoi(arg);
					h = atoi(ptr);
					if(w < 1 || h < 1)
					{
						fprintf(stderr, "invalid width, height: %d,%d. using defaults: %d,%d\n",
						       w,h,DEF_WIDTH,DEF_HEIGHT);
						cam->o.width = DEF_WIDTH;
						cam->o.height = DEF_HEIGHT;
					}
					cam->o.width = w;
					cam->o.height = h;
				}
				else
					res = -1;
				break;
			case 'G':
				if(i+1 < argc)
				{
					cam->o.gamma = atoi(argv[++i]);
				}
				else
					res = -1;
				break;
			default:
				res = -1;
				break;
			};
		}
		else
		{
			res = -1;
		}

		if(res == -1)
		{
			fprintf(stderr, "invalid option or missing parameter: %s\n", argv[i]);
			return res;
		}
	}
}

int main(int argc, char *argv[])
{
	int sockfd, clientfd;
	struct sockaddr_in remote_addr;
	int addr_len;
	pthread_t pt;
	struct caminfo *cam;
	struct connection *con;
	int res;

	cam = (struct caminfo *)malloc(sizeof(struct caminfo));
	if(!cam)
	{
		fprintf(stderr, "malloc error\n");
		exit(1);
	}

	/* ignore broken pipes n stuff */
	signal(SIGPIPE, SIG_IGN);
	signal(SIGALRM, SIG_IGN);

	/* init mutexes */
	pthread_mutex_init(&lock_queue, NULL);
	pthread_mutex_init(&lock_info, NULL);
	pthread_mutex_init(&lock_id, NULL);
	pthread_mutex_init(&lock_log, NULL);
	sem_init(&sem_con, 0, 0);

	/* defaults */
	cam->o.caption = DEF_CAPTION;
	cam->o.devfile = DEF_CAMDEVFILE;
	cam->o.port = DEF_PORT;
	cam->o.jpeg_quality = DEF_QUALITY;
	cam->o.retry_init = DEF_RETRYINIT;
	cam->o.daemon = DEF_DAEMON;
	cam->o.swap_rgb = DEF_SWAPRGB;
	cam->o.test_fps = DEF_TESTFPS;
	cam->o.width = DEF_WIDTH;
	cam->o.height = DEF_HEIGHT;
	cam->o.use_ng = DEF_USENG;
	cam->o.logfile = DEF_LOGFILE;
	cam->o.flip_horiz = DEF_FLIP_HORIZ;
	cam->o.flip_vert = DEF_FLIP_VERT;
	cam->o.gamma = DEF_GAMMA;
	cam->o.text_fg = DEF_TEXT_FG;
	cam->o.text_bg = DEF_TEXT_BG;
	cam->o.rotate = DEF_ROTATE;

	/* overwrite defaults with command line args */
	if(parse_args(cam, argc, argv) < 0)
		exit(1);

	/* don't do this stuff if we're in test mode */
	if(!cam->o.test_fps)
	{
		/* create listening socket */
		sockfd = create_and_listen(cam->o.port);
		if(sockfd < 1)
			exit(1);

		/* go into daemon mode if told to */	
		if(cam->o.daemon)
			if(daemon(0,0) < 0)
			{
				fprintf(stderr, "daemon failed.\n");
				exit(1);
			}
	}
	
	/* open camera device...retry infinitely if told to */
	if(cam->o.use_ng)
		res = ng_grab_init(cam);
	else
		res = open_cam(cam, cam->o.devfile);
	while(res < 0 && cam->o.retry_init)
	{
		usleep(5000000);
		fprintf(stderr, "camera init failed, retrying...\n");
		if(cam->o.use_ng)
			res = ng_grab_init(cam);
		else
			res = open_cam(cam, cam->o.devfile);
	}
	if(res < 0)
	{
		fprintf(stderr, "unable to initialise camera\n");
		exit(1);
	}

	if(!cam->o.use_ng)
	{
		/* get camera defaults */
		get_cam_info(cam);
		cam->vid_win.width = cam->o.width;
		cam->vid_win.height = cam->o.height;
		if(set_cam_info(cam) < 0)
		{
			fprintf(stderr, "error setting video device parameters, using defaults\n");
			get_cam_info(cam);
			if(set_cam_info(cam) < 0)
			{
				fprintf(stderr, "not a valid video device? quitting.\n");
				exit(1);
			}
		}
		cam->o.width = cam->vid_win.width;
		cam->o.height = cam->vid_win.height;
	}
	
	cam->raw_image = NULL;
	cam->grab_thread_alive = 1;
	pthread_create(&cam->pt_grab, NULL, (void*)grab_thread, (void *)cam);
	
	/* start the watchdog thread */
	pthread_create(&pt, NULL, (void*)watchdog_thread, (void *)cam);
	pthread_detach(pt);

	if(cam->o.test_fps)
	{
		pthread_join(cam->pt_grab, NULL);
		exit(0);
	}
	else
	{
		pthread_detach(cam->pt_grab);
	}

	log(cam, "webcam_server started\n");

	while((clientfd = accept(sockfd, (struct sockaddr *)&remote_addr, &addr_len)) > 0)
	{
		/* con is free'd at end of handle_connection */
		con = (struct connection*)malloc(sizeof(struct connection));
		con->cam = cam;

		/* get connection info */
		con->socketfd = clientfd;
		memcpy(&con->remote_addr, &remote_addr, addr_len);

		/* handle the connection */
		pthread_create(&pt, NULL, (void*)handle_connection, (void*)con);
		pthread_detach(pt);
	}
	
	close(sockfd);
	return 0;
}
